/*
  文字列入力関数ライブラリ
  Ver0.02
  2024.6.21

    1行入力を実装するための内部表現ライブラリ
    1行の内容に相当する文字列の保持、文字巣や表示幅などの情報取得、
    文字の追加削除などの機能を持つ
    実際の画面表示やキー入力などは上位層での実装を想定しており、
    本ライブラリでは提供しない。
    このため、本ライブラリ自体は特定の環境に依存しない

    TODO:日本語を含めた入力を想定し自作ライブラリのjstr, roman_convに
    依存する。
    用途によっては1バイト文字入力のみでよいので、
    これらのライブラリに依存しないサブセット版も提供することを検討
*/
#ifndef INPUT_LINE_H
#define INPUT_LINE_H
#include "jstr.h"
//#include <roman_conv.h>

#define PROMPT_WIDTH_MAX 64

int input_line_init(int width_max);
char *input_line_get_line(void);
int input_line_get_input(char *input);
int input_line_get_byte(void);
int input_line_get_width(void);
int input_line_get_chr_num(void);
int input_line_get_word_num(void);
int input_line_is_tail_spc(void);
int input_line_clear(void);

#ifdef DEBUG
void input_line_print_line(void);
#endif

int input_line_backspace(void);
int input_line_delete_word(void);
int input_line_append_chr(char chr);
int input_line_append_code(unsigned int code, int byte);
int input_line_set_str(char *str);
int input_line_append_str(char *str);

int input_line_set_prompt(char *new_prompt);
int input_line_set_cursor(char cursor);

#endif
