#ifndef LCD_H
#define LCD_H

#include <EzGraph.h>

#define BODY_BORDER_SIZE 8
#define SCREEN_BORDER_SIZE 8
#define SCREEN_MARGIN_SIZE 4
#define BODY_LEFT_WIDTH 160
#define BODY_RIGHT_WIDTH BODY_LEFT_WIDTH
#define BODY_TOP_HEIGHT 72
#define BODY_BOTTOM_HEIGHT 48
#define BUTTON_R 20

typedef struct {
    int x, y;    // 左上隅座標
    char **xpm;  // XPM配列へのポインタ
} LCD_fragment_info_t;

typedef struct {
    int x, y;    // 左上隅座標
    EzImage img; // ReadXPM後の画像
    int is_visible; // 非0なら表示
} LCD_fragment_t;

typedef struct LCD_unit_st {
    int type;
    int state;
    struct LCD_unit_st *next;
    //    void(*init)(struct LCD_unit_st *);
    //    void(*final)(struct LCD_unit_st *);
    void(*on_erase)(struct LCD_unit_st *);
    void(*on_act)(struct LCD_unit_st *);
    void(*on_draw)(struct LCD_unit_st *);
} LCD_unit_t;

typedef enum {
  LCD_button_down, LCD_button_left,
  LCD_button_right, LCD_button_up,
  LCD_button_a, LCD_button_b,
  LCD_button_end
} LCD_button_t;



#define LCD_TIMER_MS 50

extern LCD_fragment_info_t LCD_fragment_infos[];

int LCD_init(int width, int height);
int LCD_finalize(void);
LCD_unit_t *LCD_create_unit(int size);
void LCD_show_fragment(int fragment_no);
void LCD_hide_fragment(int fragment_no);
int LCD_is_button_down(int no);
int LCD_is_button_pushed(int no);
void LCD_set_bg(char **xpm);
void LCD_set_draw_score_handler(void(*f)(int, int));

#define LCD_CreateUnit(type) ((type*)LCD_create_unit(sizeof(type)))

// 参照しない引数についてbccが「一度も使用されない」警告を出すので
// このマクロで明示的に参照したことにする
#ifndef Ignore
#define Ignore(x) ((void)x)
#endif






#endif // LCD_H
