#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <libgen.h>
#include <sys/stat.h>
#include <string.h>
#include <EzGraph.h>
#include "layout.h"
#include "process_xpm.h"
#include "input_line.h"
#include "getarg.h"
#include "file.h"

// 切り出した2値画像に使う色とシンボル文字
#define FRAGMENT_BCOLOR "None"
#define FRAGMENT_FCOLOR "#404040"
#define FRAGMENT_BCHR "."
#define FRAGMENT_FCHR "@"

// 透過色設定するピクセルの座標
// TODO:実行時オプションで指定できるようにする
#define TRANSPRENT_X_DEFAULT 0
#define TRANSPRENT_Y_DEFAULT 0

#define FRAGMENT_NAME_MAX 16

#define FRAGMENT_NAME_WIDTH FRAGMENT_NAME_MAX

static EzImage img_main;

char **xpm_main;
dynamic_xpm_t *dxpm_main;

static int win_width  = WIN_WIDTH_DEFAULT;
static int win_height = WIN_HEIGHT_DEFAULT;
static int img_main_lx = IMG_MAIN_LX_DEFAULT;
static int img_main_ty = IMG_MAIN_TY_DEFAULT;
static int img_main_rx;
static int img_main_by;
static int transparent_x = TRANSPRENT_X_DEFAULT;
static int transparent_y = TRANSPRENT_Y_DEFAULT;

static int pixel_val2fragment_no[1 << 14];
#define PixelVal2FragmentNo(val) (pixel_val2fragment_no[val])

char **xpm_buffer;
char **xpm_icon_buffer;

char src_dirname[FILENAME_MAX];
char src_basename_noext[FILENAME_MAX];

typedef enum {
              fns_blank,  // 未記入(デフォルト)
              fns_set,    // 設定済
} fragment_name_status_t;


#define SetCurrentFragmentColor() EzSetColorByRGB(255, 0, 0)
#define SetNoncurrentFragmentColor() EzSetColorByRGB(200, 200, 200)
#define SetUnnamedColor() EzSetColorByRGB(160, 160, 0)
#define SetNamedColor() EzSetColorByRGB(120, 120, 120)

#define SetInputEnabledColor() EzSetColorByRGB(0, 0, 0)
#define SetInputDisabledColor() EzSetColorByRGB(180, 180, 180)

typedef struct {
    int lx, ty;
    int rx, by;
    int width, height;
    pixel_val_t val;
    dynamic_xpm_t *dxpm;
    //    pixel_chr_t chr;
    EzImage img;
    EzImage icon;
    char *color_name; // 元画像での色名(iconで使用)
    char name[FRAGMENT_NAME_MAX];
    fragment_name_status_t name_status;
} xpm_fragment_t;

static xpm_fragment_t xpm_fragments[COLOR_NUM_MAX];
static int xpm_fragment_num = 0;

int current_fragment_no;
xpm_fragment_t *current_fragment;
// 入力モードなら真
int is_input_enabled = 0;


void init(int argc, char *argv[]);
void final(void);
void process_xpm(void);
void draw_screen(void);    
void print_xpm_fragments(void);
void set_xpm_fragments(void);
void set_xpm_fragment(int no);
void save_xpm_fragments(char *dpath);
void keyh(int key);
void save_all(void);
void save_info(void);
void load_info(void);
void generate_xpms(void);
void set_current_fragment(int no);
int main(int argc, char *argv[]){

    init(argc, argv);
    input_line_init(FRAGMENT_NAME_WIDTH);
    process_xpm();
    puts("process_xpm done.");

    load_info();
    set_current_fragment(0);
    
    EzSetKeyHandler(keyh);
        
    draw_screen();
    //    puts("draw screen.");
    EzEventLoop();
    final();
    return 0;
}

void set_current_fragment(int no){
    current_fragment_no = no;
    current_fragment = &xpm_fragments[no];
    input_line_set_str(current_fragment->name);
    printf("set_current_fragment:%d %s\n", no, current_fragment->name);
}

void inc_current_fragment(void){
    int no;
    assert(current_fragment != NULL);

    //    input_line_get_input(current_fragment->name);
    no = current_fragment_no + 1;
    if (no >= xpm_fragment_num){
        no = 0;
    }
    set_current_fragment(no);
}
void dec_current_fragment(void){
    int no;
    assert(current_fragment != NULL);

    //    input_line_get_input(current_fragment->name);
    no = current_fragment_no -1;
    if (no < 0){
        no = xpm_fragment_num - 1;
    }
    set_current_fragment(no);
}


#define CodeCtrl(c) ((c)-'a'+1)
                          
void keyh(int key){
    switch (key){
    case CodeCtrl('s'):
        save_info();
        break;
    case CodeCtrl('g'):
        generate_xpms();
        break;
    case EZ_Right:
        inc_current_fragment();
        break;
    case EZ_Left:
        dec_current_fragment();
        break;
    case 13:
        if (is_input_enabled){
            input_line_get_input(current_fragment->name);
            current_fragment->name_status = fns_set;
            is_input_enabled = 0;
            printf("|%s|\n", input_line_get_line());
        }
        else{
            is_input_enabled = -1;
        }
        break;
    case '\b':  // BackSpace 
        printf("BS!\n");
        input_line_backspace();
        break;
    default:
        if (is_input_enabled){
            if (isalnum(key) || key == '_'){
                input_line_append_chr(key);
            }
        }
        break;
    }
    draw_screen();
}

void init(int argc, char *argv[]){
    char path[FILENAME_MAX];
    char *p;

    // 引数は1個(xmpfile)固定
    register_arg_str("xpmfile");
    // オプション解析に失敗したとき
    if (getarg(argc, argv) != 0){
        log_error("オプションに誤りがあります。\n");
        exit(1);
    }
    // (オプションを除く)引数の個数に過不足があるとき
    if (get_pure_argc() != 1){
        log_error("引数を1個だけ指定してください。\n");
        opt_help(0, NULL);
        exit(1);
    }

    strcpy(path, get_pure_argv()[0]);
    strcpy(src_dirname, dirname(path));
    strcpy(path, get_pure_argv()[0]);
    strcpy(src_basename_noext, basename(path));
    // パスが拡張子を含む場合は取り除く
    if ((p = strrchr(src_basename_noext, '.')) != NULL){
        *p = '\0';
    }

    printf("|%s|%s|%s|\n", src_dirname, src_basename_noext, path);
    
    print_current_options();
    EzOpenWindow(win_width, win_height);
    
    //    EzSetMouseHandler(dummy_mouseh);
}


void final(void){
    free_xpm_buffer(xpm_main);
    free_dynamic_xpm(dxpm_main);

    free_xpm_buffer(xpm_buffer);
    free_xpm_buffer(xpm_icon_buffer);
}


void process_xpm(void){
    char *load_filename = get_pure_argv()[0];
    //    char save_filename[1024];
    
    printf("lcdgame img converter\n");
    printf("load %s\n", load_filename); 

    xpm_main = load_xpm(load_filename, NULL);
    //    printf("%p\n", xpm_main);
    //    for (i=0; i<20; i++){
    //        printf("%d:|%s|\n", i, xpm_main[i]);
    //    }

    dxpm_main = xpm2dynamic_xpm(xpm_main, NULL);
    // 指定位置(デフォルト左上隅)の色を背景色として透過色設定する
    DXpmTransparentVal(dxpm_main)
        = DXpmPixel(dxpm_main, transparent_x, transparent_y);
    set_xpm_fragments();

    //    print_xpm_fragments();
    
    //    printf("pixel(0,0):%d\n", DXpmPixel(dxpm_main, 0, 0));
    //    printf("transparentval:%d\n", DXpmTransparentVal(dxpm_main));
    DXpmTransparentVal(dxpm_main) = DXpmPixel(dxpm_main, 0, 0);
    //    printf("transparentval:%d\n", DXpmTransparentVal(dxpm_main));

    img_main = EzReadXPM(xpm_main);

    img_main_rx = img_main_lx + dxpm_main->width;
    img_main_by = img_main_ty + dxpm_main->height;

    //    sprintf(save_filename, "%s.out.xpm", load_filename);
    //    save_xpm(save_filename, "hoge", xpm_main);

    // save_xpm_fragments();
}

void set_xpm_fragments(void){
    int i, fragment_no=0;
    pixel_val_t pixel_val;
    int x, y;
    xpm_fragment_t *fragment;
    dynamic_xpm_t *dxpm;
    int fragment_width_max=0, fragment_height_max=0;
    dynamic_xpm_t *dxpm_icon;
        
    
    for (i = 0; i < DXpmColorNum(dxpm_main); i++){
        //        printf("check color %d\n", i);
        // 透過色以外の色を切り出す
        pixel_val = DXpmColorVal(dxpm_main, i);
        if (pixel_val == DXpmTransparentVal(dxpm_main)){
            continue;
        }
        xpm_fragments[fragment_no].lx = DXpmWidth(dxpm_main);
        xpm_fragments[fragment_no].rx = -1;
        xpm_fragments[fragment_no].ty = DXpmHeight(dxpm_main);
        xpm_fragments[fragment_no].by = -1;
        xpm_fragments[fragment_no].val = pixel_val;
        xpm_fragments[fragment_no].color_name
            = DXpmColorName(dxpm_main, i);
        strcpy(xpm_fragments[fragment_no].name, "");
        xpm_fragments[fragment_no].name_status = fns_blank;
        PixelVal2FragmentNo(pixel_val) = fragment_no;
        fragment_no++;
    }
    xpm_fragment_num = fragment_no;

    for (y = 0; y < DXpmHeight(dxpm_main); y++){
        for (x = 0; x < DXpmWidth(dxpm_main); x++){
            pixel_val = DXpmPixel(dxpm_main, x, y);
            if (pixel_val == DXpmTransparentVal(dxpm_main)){
                continue;
            }
            fragment_no = PixelVal2FragmentNo(pixel_val);
            fragment = &xpm_fragments[fragment_no];
            if (fragment->lx > x) {fragment->lx = x;}
            if (fragment->rx < x) {fragment->rx = x;}
            if (fragment->ty > y) {fragment->ty = y;}
            if (fragment->by < y) {fragment->by = y;}
            //            printf("!(%d, %d) %d %x\n", x, y, fragment_no, pixel_val);
        }
    }
    for (i = 0; i < xpm_fragment_num; i++){
        fragment = &xpm_fragments[i];
        fragment->width = fragment->rx - fragment->lx + 1;
        fragment->height = fragment->by - fragment->ty + 1;
        if (fragment_width_max < fragment->width){
            fragment_width_max = fragment->width;
        }
        if (fragment_height_max < fragment->height){
            fragment_height_max = fragment->height;
        }
    }
    printf("fragments width/height max:%dx%d\n",
           fragment_width_max, fragment_height_max);

    xpm_buffer = alloc_xpm_buffer(fragment_width_max,
                                  fragment_height_max, 2);
    xpm_icon_buffer = alloc_xpm_buffer(FRAGMENT_ICON_WIDTH,
                                       FRAGMENT_ICON_HEIGHT, 2);

    for (i = 0; i < xpm_fragment_num; i++){
        fragment = &xpm_fragments[i];
        fragment->dxpm
            = alloc_dynamic_xpm(fragment->width,
                                fragment->height);
        dxpm = fragment->dxpm;
        DXpmColorNum(dxpm) = 2;
        DXpmDepth(dxpm) = 1;
        DXpmColorVal(dxpm, 0)
            = DXpmTransparentVal(dxpm) 
            = PixelStr2Val(FRAGMENT_BCHR, dxpm->depth);
        strcpy(DXpmColorName(dxpm, 0), FRAGMENT_BCOLOR);
        DXpmColorVal(dxpm, 1)
            = PixelStr2Val(FRAGMENT_FCHR, dxpm->depth);
        strcpy(DXpmColorName(dxpm, 1), FRAGMENT_FCOLOR);
        for (y = 0; y < fragment->height; y++){
            for (x = 0; x < fragment->width; x++){
                if (DXpmPixel(dxpm_main,
                              x + fragment->lx, y + fragment->ty)
                    == fragment->val)
                {
                    DXpmPixel(dxpm, x, y)
                        = DXpmColorVal(dxpm, 1);
                }
                else {
                    DXpmPixel(dxpm, x, y)
                        = DXpmColorVal(dxpm, 0);
                }
            }
        }
        printf("No.%d \n", i);
        //        show_dynamic_xpm(dxpm);
        dynamic_xpm2xpm(dxpm, xpm_buffer);
        fragment->img = EzReadXPM(xpm_buffer);

        dxpm_icon = alloc_dynamic_xpm(FRAGMENT_ICON_WIDTH,
                                      FRAGMENT_ICON_HEIGHT);
        DXpmColorNum(dxpm_icon) = 2;
        DXpmDepth(dxpm_icon) = 1;
        DXpmColorVal(dxpm_icon, 0) = DXpmColorVal(dxpm, 0);
        strcpy(DXpmColorName(dxpm_icon, 0), FRAGMENT_BCOLOR);
        DXpmColorVal(dxpm_icon, 1) = DXpmColorVal(dxpm, 1);
        strcpy(DXpmColorName(dxpm_icon, 1),
               fragment->color_name);

        for (y = 0; y < FRAGMENT_ICON_HEIGHT; y++){
            for (x = 0; x < FRAGMENT_ICON_WIDTH; x++){

                if (x < DXpmWidth(dxpm) &&
                    y < DXpmHeight(dxpm))
                {
                    //                    printf("dx:(%d,%d) %d %d %d %d %d\n", x, y, DXpmPixel(dxpm, x, y), DXpmColorVal(dxpm, 0), DXpmColorVal(dxpm, 1),
                    //                           DXpmColorVal(dxpm_icon, 0), DXpmColorVal(dxpm_icon, 1));
                    DXpmPixel(dxpm_icon, x, y)

                        = DXpmPixel(dxpm, x, y);
                }
                else {
                    DXpmPixel(dxpm_icon, x, y)
                        = DXpmColorVal(dxpm, 0);
                }
            }
        }

        dynamic_xpm2xpm(dxpm_icon, xpm_icon_buffer);
        show_xpm(xpm_buffer);
        show_xpm(xpm_icon_buffer);
        fragment->icon = EzReadXPM(xpm_icon_buffer);
    }
    set_current_fragment(0);
}

void draw_screen(void){
    int i;
    char buf[16];
    int flx = FRAGMENT_LX_DEFAULT;
    int fty = img_main_by + MARGIN;

    EzSetColorByRGB(255, 255, 255);
    EzFillBoxB(0, 0, win_width-1, win_height-1);
    EzPutB(img_main_lx, img_main_ty, img_main);
    //    EzSetColorByRGB(255, 255, 255);
    EzSetColorByRGB(0, 0, 0);
    EzDrawBoxB(img_main_lx-1, img_main_ty-1,
               img_main_rx+1, img_main_by+1);

    EzSetFontSize(PREVIEW_FONTSIZE);
    for (i = 0; i < xpm_fragment_num; i++){
        if (i == current_fragment_no){
            SetCurrentFragmentColor();
        }
        else {
            SetNoncurrentFragmentColor();
        }
        EzDrawBoxB(xpm_fragments[i].lx+img_main_lx, 
                  xpm_fragments[i].ty+img_main_ty, 
                  xpm_fragments[i].rx+img_main_lx, 
                  xpm_fragments[i].by+img_main_ty);

        if (xpm_fragments[i].name_status == fns_blank){
            SetUnnamedColor();
        }
        else {
            SetNamedColor();
        }
        sprintf(buf, "%02d", i);
        EzDrawStringB(xpm_fragments[i].lx+img_main_lx,
                      xpm_fragments[i].ty+img_main_ty,
                      buf);
        //        printf("put i:%d %dx%d \n", i,
        //       xpm_fragments[i].width,xpm_fragments[i].height);
        //        EzPutB(48, i*FRAGMENT_ICON_LINESKIP, xpm_fragments[i].imgc);
        EzPutB(0, i*FRAGMENT_ICON_LINESKIP, xpm_fragments[i].icon);
    }
    EzSetFontSize(FRAGMENT_FONTSIZE);
    sprintf(buf, "%d/%d", current_fragment_no,
            xpm_fragment_num);
    EzDrawStringB(flx, fty+FRAGMENT_FONTSIZE, buf);
    if (is_input_enabled){
        SetInputEnabledColor();
    }
    else {
        SetInputDisabledColor();
    }
    EzDrawStringB(flx, fty+FRAGMENT_FONTSIZE*2,
                  input_line_get_line());
    EzPutB(flx, fty+FRAGMENT_FONTSIZE*3,
           xpm_fragments[current_fragment_no].img);
    
    EzShowBuffer();
    printf("drawn_screen|%s|\n", input_line_get_line());
}


void print_xpm_fragments(void){
    int i;
    for (i = 0; i < xpm_fragment_num; i++){
        printf("%3d %d (%d, %d)-(%d, %d) (%dx%d) \n",
               //        printf("%3d %c%c (%d, %d) %dx%d \n",
               i,
               //               '@','@',
               //               xpm_fragments[i].str.chr[0],
               //               xpm_fragments[i].str.chr[1],
               xpm_fragments[i].val, 
               xpm_fragments[i].lx, 
               xpm_fragments[i].ty, 
               xpm_fragments[i].rx, 
               xpm_fragments[i].by, 
               xpm_fragments[i].width, 
               xpm_fragments[i].height);
    }
}


void save_xpm_fragments(char *dpath){
    FILE *hfp, *cfp;
    int i;
    char save_fpath[FILENAME_MAX];
    xpm_fragment_t *fragment;
    char xpm_name[FILENAME_MAX];
    char cheader_fpath[FILENAME_MAX];
    char csrc_fpath[FILENAME_MAX];
    
    sprintf(cheader_fpath, "%s.h", dpath);
    hfp = fopen(cheader_fpath, "w");
    if (hfp == NULL) {
        printf("ヘッダファイル'%s'の生成に失敗しました。\n",
               cheader_fpath);
        exit(1);
    }
    sprintf(csrc_fpath, "%s.c", dpath);
    cfp = fopen(csrc_fpath, "w");
    if (cfp == NULL) {
        printf("ソースファイル'%s'の生成に失敗しました。\n",
               csrc_fpath);
        exit(1);
    }

    fprintf(hfp, "#define LCD_FRAGMENT_NUM %d\n",
            xpm_fragment_num);
    fprintf(cfp, "#include<EzGraph.h>\n");
    fprintf(cfp, "#include\"lcd.h\"\n");
    fprintf(cfp, "#include \"%s\"\n\n", cheader_fpath);
    
    for (i = 0; i < xpm_fragment_num; i++){
        fragment = &xpm_fragments[i];
        if (fragment->name_status == fns_set){
            sprintf(save_fpath, "%s/%s.xpm",
                    dpath, fragment->name);
            sprintf(xpm_name, "%s_xpm", fragment->name);
        }
        else {
            sprintf(save_fpath, "%s/%03d.xpm",
                    dpath, i);
            sprintf(xpm_name, "%03d_xpm", i);
        }
        printf("saving '%s'\n", save_fpath);
        dynamic_xpm2xpm(fragment->dxpm, xpm_buffer);
        save_xpm(save_fpath, xpm_name, xpm_buffer);        
        fprintf(cfp, "#include\"");
        fputs(save_fpath, cfp);
        fprintf(cfp, "\"\n");       
    }
    fprintf(cfp, "LCD_fragment_info_t "
            "LCD_fragment_infos[] = {\n");
    for (i = 0; i < xpm_fragment_num; i++){
        fragment = &xpm_fragments[i];
        if (fragment->name_status == fns_set){
            sprintf(xpm_name, "%s_xpm", fragment->name);
        }
        else {
            sprintf(xpm_name, "%03d_xpm", i);
        }
        fprintf(cfp, "    {%d, %d, %s},\n",
                fragment->lx, fragment->ty, xpm_name);
    }
    fprintf(cfp, "};\n");

    fprintf(hfp, "typedef enum {\n");
    for (i = 0; i < xpm_fragment_num; i++){
        fragment = &xpm_fragments[i];
        if (fragment->name_status == fns_set){
            fprintf(hfp, "    LCD_fn_%s,\n", fragment->name);
        }
        else {
            fprintf(hfp, "    LCD_fn_%03d\n", i);
        }
    }
    fprintf(hfp, "} LCD_fragment_name_t;\n");

    fprintf(cfp, "int LCD_get_fragment_num(void){\n");
    fprintf(cfp, "    return LCD_FRAGMENT_NUM;\n");
    fprintf(cfp, "}\n");
    
    fclose(hfp);
    fclose(cfp);
}

//void save_all(void){
//    printf("Saving all data...\n");
//    save_info();
//}

void load_info(void){
    int i;
    char path[FILENAME_MAX];
    FILE *fp;
    xpm_fragment_t *fragment;
    int lx, ty, width, height;
    char line[LINE_MAX];
    char name[FRAGMENT_NAME_MAX];
    
    sprintf(path, "%s/%s.dat", src_dirname, src_basename_noext);
    if (file_if_exists(path)) {
        printf("loading info '%s'\n", path);
    }
    else {
        printf("No file '%s'\n", path);
        return;
    }
    
    fp = fopen(path, "r");
    if (fp == NULL){
        log_error("ファイル%sを読み込みオープンできません");
        exit(1);
    }

    while (fgets(line, LINE_MAX, fp) != NULL){
        sscanf(line, "%d:(%d,%d):%dx%d:%s\n",
               &i, &lx, &ty, &width ,&height, name);
        printf("read %d:(%d,%d) %dx%d:%s\n",
               i, lx, ty, width, height, name);
        fragment = &xpm_fragments[i];
        if (name[0] != '-' &&
            fragment->lx == lx &&
            fragment->ty == ty &&
            fragment->width == width &&
            fragment->height == height)
        {
            printf("  copy name %s\n", name);
            strcpy(fragment->name, name);
            fragment->name_status = fns_set;
        }
    }
    fclose(fp);
}


void save_info(void){
    int i;
    char path[FILENAME_MAX];
    FILE *fp;
    xpm_fragment_t *fragment;
    
    sprintf(path, "%s/%s.dat", src_dirname, src_basename_noext);
    printf("Saving info in '%s'\n", path);

    fp = fopen(path, "w");
    if (fp == NULL){
        log_error("ファイル%sを書き込みオープンできません");
        exit(1);
    }
    
    for (i = 0; i < xpm_fragment_num; i++){
        fragment = &xpm_fragments[i];
        fprintf(fp, "%d:(%d,%d):%dx%d:", i,
                fragment->lx, fragment->ty,
                fragment->width, fragment->height);
        if (fragment->name_status == fns_set){
            fprintf(fp, "%s\n", fragment->name);
        }
        else {
            fprintf(fp, "-\n");
        }
    }
    fclose(fp);
}

void generate_xpms(void){
    struct stat stat_buf;
    int stat_ret;
    char out_dpath[FILENAME_MAX];
    
    printf("XPMを生成します...\n");
    strncpy(out_dpath, src_basename_noext, FILENAME_MAX-64);
    printf("出力先ディレクトリ'%s'を確認します。\n", out_dpath);
    stat_ret = stat(out_dpath, &stat_buf);
    if (stat_ret != 0) { // ファイルが存在しない
        if (mkdir(out_dpath, 0755) == 0){
            printf("'%s'を作成しました。\n", out_dpath);
        }
        else {
            printf("'%s'の作成に失敗しました。\n", out_dpath);
            printf("XPM生成処理を中止します。\n");
            exit(1);
        }
    }
    else {
        if (S_ISDIR(stat_buf.st_mode)){
            printf("既存の'%s'に上書きします。\n", out_dpath);
        }
        else {
            printf("ディレクトリではないファイル'%s'が存在します。", out_dpath);
            printf("XPM生成処理を中止します。\n");
            exit(1);
        }
    }

    save_xpm_fragments(out_dpath);
}
