#include <EzGraph.h>
#include <stdio.h>
#include <time.h>
#include "rnd.h"
#include "lcd.h"
#include "tree3.h"
#include "tree_game.h"
#include "sample/tree_bg.xpm"

void init(void);
void reset_game(void);
void set_game_state(game_state_t new_state);
void manager_act(LCD_unit_t *unit);
void manager_erase(LCD_unit_t *unit);
void manager_draw(LCD_unit_t *unit);
void chee_act(LCD_unit_t *unit);
void chee_erase(LCD_unit_t *unit);
void chee_draw(LCD_unit_t *unit);
void gondin_act(LCD_unit_t *unit);
void gondin_erase(LCD_unit_t *unit);
void gondin_draw(LCD_unit_t *unit);
void gondin_spawn(unit_gondin_t *gondin);
void inc_miss(void);
void add_score(int val);
void set_level(int val);
void inc_level(void);
void add_exp(int val);

void draw_score(int client_left, int client_top);



typedef struct {
    LCD_fragment_name_t body, wingd, wingu;
} chee_fragment_t;

chee_fragment_t chee_fragments[4] =
{
 {LCD_fn_CheeLF, LCD_fn_WingDLF, LCD_fn_WingULF},
 {LCD_fn_CheeLN, LCD_fn_WingDLN, LCD_fn_WingULN},
 {LCD_fn_CheeRN, LCD_fn_WingDRN, LCD_fn_WingURN},
 {LCD_fn_CheeRF, LCD_fn_WingDRF, LCD_fn_WingURF},
};

LCD_fragment_name_t gondin_fragments[4][6] =
{
 {LCD_fn_GondinLF0,LCD_fn_GondinLF1,LCD_fn_GondinLF2,
  LCD_fn_GondinLF3,LCD_fn_GondinLF4,LCD_fn_GondinLF5},
 {0, LCD_fn_GondinLN1,LCD_fn_GondinLN2,
  LCD_fn_GondinLN3,LCD_fn_GondinLN4,LCD_fn_GondinLN5},
 {0, LCD_fn_GondinRN1,LCD_fn_GondinRN2,
  LCD_fn_GondinRN3,LCD_fn_GondinRN4,LCD_fn_GondinRN5},
 {LCD_fn_GondinRF0,LCD_fn_GondinRF1,LCD_fn_GondinRF2,
  LCD_fn_GondinRF3,LCD_fn_GondinRF4,LCD_fn_GondinRF5},
};

LCD_fragment_name_t miss_fragments[3] = {
    LCD_fn_Miss1,
    LCD_fn_Miss2,
    LCD_fn_Miss3
};


game_state_t game_state;
game_t game;
unit_manager_t *manager;
unit_chee_t *chee;
unit_gondin_t *gondins[GONDIN_MAX];

int main(void){
    LCD_init(720, 480);
    LCD_set_bg(tree_bg);
    init();
    EzEventLoop();    
    LCD_finalize();
    return 0;
}

void init(void){
    int i;

    srnd((unsigned int)time(NULL));
    
    manager = LCD_CreateUnit(unit_manager_t);
    manager->unit.type = ut_manager;
    manager->unit.state = us_chee_normal,
    manager->unit.on_act = manager_act;
    manager->unit.on_erase = manager_erase;
    manager->unit.on_draw = manager_draw;

    chee = LCD_CreateUnit(unit_chee_t);
    chee->unit.type = ut_chee;
    chee->unit.state = 0; //!!!
    chee->unit.on_act = chee_act;
    chee->unit.on_erase = chee_erase;
    chee->unit.on_draw = chee_draw;
    chee->x = 1;
    for (i = 0; i < GONDIN_MAX; i++){
        gondins[i] = LCD_CreateUnit(unit_gondin_t);
        gondins[i]->unit.type = ut_gondin;
        gondins[i]->unit.state = 0; //!!!
        gondins[i]->unit.on_act = gondin_act;
        gondins[i]->unit.on_erase = gondin_erase;
        gondins[i]->unit.on_draw = gondin_draw;
        gondins[i]->is_alive = 0;
    }
    set_game_state(gs_title);
    LCD_set_draw_score_handler(draw_score);
}


void reset_game(void){
    int i;

    game.score = 0;
    game.miss = 0;
    set_level(0);
    game.exp = 0;
    chee->unit.state = us_chee_normal;
    chee->x = 1;
    for (i = 0; i < GONDIN_MAX; i++){
        gondins[i]->is_alive = 0;
    }
}   

void set_game_state(game_state_t new_state){
    switch (new_state){
    case gs_title:
        manager->step_ctr = 0;
        break;
    case gs_playing:
        manager->step_ctr = 0;
        reset_game();
        break;
    case gs_gameover:
        manager->step_ctr = GAMEOVER_WAIT;
        break;
    default:
        break;
    }
    game_state = new_state;
}

void inc_miss(void){
    game.miss++;
    if (game.miss == GAME_MISS_LIMIT){
        set_game_state(gs_gameover);
    }
}

void add_score(int val){
    game.score += val;
    if (game.score > GAME_SCORE_LIMIT){
        game.score = GAME_SCORE_LIMIT;
    }
}
void set_level(int val){
    int respawn_level;
    
    game.level = val;
    game.current_gondin_max
        = game.level / (GONDIN_RESPAWN_LEVEL_MAX+1) + 1;
    respawn_level
        = game.level % (GONDIN_RESPAWN_LEVEL_MAX+1);
    game.current_respawn_wait
        = GONDIN_RESPAWN_BASEWAIT
        - GONDIN_RESPAWN_DEC * respawn_level;
}    

void inc_level(void){
    int level = game.level + 1;
    if (level > GAME_LEVEL_MAX){
        level = 0;
    }
    set_level(level);
    printf("GAME LEVEL %d/%d\n", game.level, GAME_LEVEL_MAX);
}

void add_exp(int val){
    game.exp += val;
    if (game.exp >= GAME_LEVELUP_EXP){
        inc_level();
        game.exp = 0;
    }
}

void manager_act(LCD_unit_t *unit){
    unit_manager_t *manager = (unit_manager_t *)unit;
    int i;
    printf("act manager gamestate:%d ctr:%d [score:%d miss:%d] [Lv%d(%d/%d) (maxgondin:%d wait:%d]\n", game_state, manager->step_ctr, game.score, game.miss, game.level, game.exp, GAME_LEVELUP_EXP, game.current_gondin_max, game.current_respawn_wait);

    switch (game_state){
    case gs_title:
        if (LCD_is_button_down(LCD_button_a)){
            set_game_state(gs_playing);
        }
        break;
    case gs_playing:
        manager->step_ctr--;
        if (manager->step_ctr <= 0){
            for (i = 0 ; i < game.current_gondin_max; i++){
                if (!gondins[i]->is_alive){
                    gondin_spawn(gondins[i]);
                    manager->step_ctr
                        = game.current_respawn_wait
                        + RRnd(-30, 30);
                    break;
                }
            }
        }
        // 暫定的にBボタンでゲームを強制終了できる
        if (LCD_is_button_down(LCD_button_b)){
            set_game_state(gs_title);
        }
        break;
    case gs_gameover:
        manager->step_ctr--;
        if (manager->step_ctr <= 0){
            set_game_state(gs_title);
        }
        break;
    default:
        break;
    }
}

void manager_erase(LCD_unit_t *unit){
    unit_manager_t *manager = (unit_manager_t *)unit;
    int i;
    Ignore(manager);

    // ゲーム中以外はとりあえず何もしない
    if (game_state == gs_title) return;

    for (i = 0; i < GAME_MISS_LIMIT; i++){
        LCD_hide_fragment(miss_fragments[i]);
    }
}
void manager_draw(LCD_unit_t *unit){
    unit_manager_t *manager = (unit_manager_t *)unit;
    int i;
    Ignore(manager);

    // ゲーム中以外はとりあえず何もしない
    if (game_state == gs_title) return;

    for (i = 0; i < game.miss; i++){
        LCD_show_fragment(miss_fragments[i]);
    }
}

void chee_act(LCD_unit_t *unit){
    unit_chee_t *chee = (unit_chee_t *)unit;

    // ゲーム中以外はとりあえず何もしない
    if (game_state == gs_title) return;
    
    if (LCD_is_button_pushed(LCD_button_left) && chee->x > 0){
        chee->unit.state = us_chee_normal;
        chee->x--;
    }
    if (LCD_is_button_pushed(LCD_button_right) && chee->x < 3){
        chee->unit.state = us_chee_normal;
        chee->x++;
    }
    //    printf("acting chee. %d %d\n", LCD_is_button_pushed(LCD_button_left),LCD_is_button_pushed(LCD_button_right));
}
void chee_erase(LCD_unit_t *unit){
    unit_chee_t *chee = (unit_chee_t *)unit;

    // ゲーム中以外はとりあえず何もしない
    if (game_state == gs_title) return;

    LCD_hide_fragment(chee_fragments[chee->x].body);
    LCD_hide_fragment(chee_fragments[chee->x].wingd);
    LCD_hide_fragment(chee_fragments[chee->x].wingu);
}
void chee_draw(LCD_unit_t *unit){
    unit_chee_t *chee = (unit_chee_t *)unit;

    // ゲーム中以外はとりあえず何もしない
    if (game_state == gs_title) return;

    LCD_show_fragment(chee_fragments[chee->x].body);
    if (chee->unit.state == us_chee_catching){
        LCD_show_fragment(chee_fragments[chee->x].wingu);
    }
    else {
        LCD_show_fragment(chee_fragments[chee->x].wingd);
    }
}

void gondin_spawn(unit_gondin_t *gondin){
    int i;
    int is_conflict;
    printf("new gondin.\n");

    gondin->is_alive = 1;
    gondin->fall_ctr = 0;
    do {
        gondin->x = RRnd(0, 3);
        if (gondin->x == 1 || gondin->x == 2){
            gondin->y = 1;
        }
        else {
            gondin->y = 0;
        }
        is_conflict = 0;
        for (i = 0; i < game.current_gondin_max; i++){
            if (gondins[i] == gondin) break;
            if (gondins[i]->x == gondin->x &&
                gondins[i]->y == gondin->y)
            {
                is_conflict = 1;
            }
        }
    } while (is_conflict);
}

void gondin_act(LCD_unit_t *unit){
    unit_gondin_t *gondin = (unit_gondin_t *)unit;
    Ignore(gondin);

    // ゲーム中以外はとりあえず何もしない
    if (game_state == gs_title) return;

    // 生きているとき
    if (gondin->is_alive){
        if (gondin->y == 4 && chee->x == gondin->x){
            chee->unit.state = us_chee_catching;
            add_score(1);
            add_exp(1);
            gondin->is_alive = 0;
        }
        gondin->fall_ctr++;
        if (gondin->fall_ctr >= GONDIN_FALL_WAIT){
            gondin->y++;
            gondin->fall_ctr = 0;
            if (gondin->y == 6){ // 破片表示時間終了
                gondin->is_alive = 0;
            }
            else if (gondin->y == 5){
                inc_miss();
            }
        }
    }
}


void gondin_erase(LCD_unit_t *unit){
    unit_gondin_t *gondin = (unit_gondin_t *)unit;

    // ゲーム中以外はとりあえず何もしない
    if (game_state == gs_title) return;
    if (!gondin->is_alive) return;
    LCD_hide_fragment(gondin_fragments[gondin->x][gondin->y]);    
}

void gondin_draw(LCD_unit_t *unit){
    unit_gondin_t *gondin = (unit_gondin_t *)unit;

    // ゲーム中以外はとりあえず何もしない
    if (game_state == gs_title) return;
    if (!gondin->is_alive) return;
    //    puts("show gondin!");
    LCD_show_fragment(gondin_fragments[gondin->x][gondin->y]);   }

void draw_score(int client_left, int client_top){
    if (game_state == gs_title) return;
    char buf[5];

    sprintf(buf, "%04d", game.score);
    EzSetFontSize(48);
    EzSetColorByRGB(40, 40, 40);
    EzDrawStringB(client_left+440, client_top+48, buf);    
}
